﻿using Microsoft.AspNetCore.Mvc;
using System.Globalization;
using System.Linq;
using System.Linq.Dynamic.Core;
using System.Reflection;
using System.Text;

namespace ComponentPractice.Services
{
    public class ToCSVService
    {
        public IQueryable ApplyQuery<T>(IQueryable<T> items, IQueryCollection query = null) where T: class
        {
            if (query == null)
                return items;

            //// For EF
            //if (query.ContainsKey("$expand"))
            //{
            //    var propertiesToExpand = query["$expand"].ToString().Split(',');
            //    foreach (var p in propertiesToExpand)
            //    {
            //        items = items.Include(p);
            //    }
            //}

            var filter = query.ContainsKey("$filter") ? query["$filter"].ToString() : null;
            if (!string.IsNullOrEmpty(filter))
            {
                items = items.Where(filter);
            }

            if (query.ContainsKey("$orderBy"))
            {
                items = items.OrderBy(query["$orderBy"].ToString());
            }

            if (query.ContainsKey("$skip"))
            {
                items = items.Skip(int.Parse(query["$skip"].ToString()));
            }

            if (query.ContainsKey("$top"))
            {
                items = items.Take(int.Parse(query["$top"].ToString()));
            }

            if (query.ContainsKey("$select"))
            {
                return items.Select($"new ({query["$select"].ToString()})");
            }

            return items;
        }

        public FileStreamResult ToCSV(IQueryable query, string fileName = null)
        {
            var columns = GetProperties(query.ElementType);
            
            var sb = new StringBuilder();
            foreach (var item in query)
            {
                var row = new List<string>();
                foreach (var col in columns)
                {
                    if (col.Value == typeof(DateTime))
                        row.Add(((DateTime)GetValue(item, col.Key)).ToString("yyyy-MM-dd HH:mm:ss", CultureInfo.InvariantCulture));
                    else
                        row.Add($"{GetValue(item, col.Key)}".Trim());
                }

                sb.AppendLine(string.Join(",", row.ToArray()));
            }

            var result = new FileStreamResult(new MemoryStream(Encoding.UTF8.GetBytes($"{string.Join(",", columns.Select(c => c.Key))}{System.Environment.NewLine}{sb.ToString()}")), "text/csv");
            result.FileDownloadName = $"{(!string.IsNullOrEmpty(fileName) ? fileName : "Export")}.csv";

            return result;
        }

        public static object GetValue(object target, string name)
        {
            return target.GetType().GetProperty(name).GetValue(target);
        }

        public static IEnumerable<KeyValuePair<string, Type>> GetProperties(Type type)
        {
            return type.GetProperties(BindingFlags.Public | BindingFlags.Instance)
                .Where(p => p.CanRead && IsSimpleType(p.PropertyType)).Select(p => new KeyValuePair<string, Type>(p.Name, p.PropertyType));
        }

        public static bool IsSimpleType(Type type)
        {
            var underlyingType = type.IsGenericType &&
                type.GetGenericTypeDefinition() == typeof(Nullable<>) ?
                Nullable.GetUnderlyingType(type) : type;

            if (underlyingType == typeof(System.Guid))
                return true;

            var typeCode = Type.GetTypeCode(underlyingType);

            switch (typeCode)
            {
                case TypeCode.Boolean:
                case TypeCode.Byte:
                case TypeCode.Char:
                case TypeCode.DateTime:
                case TypeCode.Decimal:
                case TypeCode.Double:
                case TypeCode.Int16:
                case TypeCode.Int32:
                case TypeCode.Int64:
                case TypeCode.SByte:
                case TypeCode.Single:
                case TypeCode.String:
                case TypeCode.UInt16:
                case TypeCode.UInt32:
                case TypeCode.UInt64:
                    return true;
                default:
                    return false;
            }
        }
    }
}
